//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "common.h"

//---------------------------------------------------------------------------
// privatne funkcie
int processArguments(int argc, char* argv[], WORK_PARAMS *pWP);
int parseNumber(char *pszStr);
void version(void);
void usage(void);

//---------------------------------------------------------------------------
int main(int argc, char* argv[])
{
  WORK_PARAMS *pWP = NULL;
  int iErr;

  do {
    if (argc == 1) {
      iErr = ERR_PRINT_USAGE;
      break;
    }

    // alokovanie pamati pre pracovne premenne
    iErr = allocBuffer((void *) &pWP, sizeof(WORK_PARAMS));
    if (iErr != ERR_OK)
      break;

    // spracovanie a kontrola vstupnych parametrov
    iErr = processArguments(argc, argv, pWP);
    if (iErr != ERR_OK)
      break;

    // alokovanie pamati pre pracovny buffer
    iErr = allocBuffer((void *) &pWP->pbtBuffer, 65600);
    if (iErr != ERR_OK)
      break;

    // precitanie vstupneho suboru pasky a priprava zoznamu blokov
    iErr = readTapeFile(pWP);
    if (iErr != ERR_OK)
      break;

    // vsetky bloky, ak nebolo zadane
    if (pWP->iBlockFrom < 0 || pWP->iBlockTo < 0) {
      pWP->iBlockFrom = 1;
      pWP->iBlockTo = pWP->iBlockCount;
    }

    // zobrazenie zoznamu blokov na paske
    if (pWP->bList) {
      printBlockList(pWP);
      break;
    }

    // vytvorenie WAV suboru
    iErr = createWav(pWP);

  } while (FALSE);

  if (iErr == ERR_PRINT_USAGE) {
    usage();
    iErr = ERR_OK;
  }
  else if (iErr == ERR_PRINT_VERSION) {
    version();
    iErr = ERR_OK;
  }
  else if (iErr == ERR_OK)
    printf("All done.\n");

  if (pWP) {
    if (pWP->fOut)
      fclose(pWP->fOut);

    // uvolnenie alokovanej pamati
    freeAllBlocks(pWP->pBlocks);
    if (pWP->pbtBuffer)
      free(pWP->pbtBuffer);
    if (pWP->pbtLevels)
      free(pWP->pbtLevels);
    if (pWP->pWH)
      free(pWP->pWH);

    free(pWP);
  }

#ifdef _DEBUG
  getchar();
#endif

  return -iErr;
}
//---------------------------------------------------------------------------
int processArguments(int argc, char* argv[], WORK_PARAMS *pWP)
{
  int iErr = 0, i, iV;
  char *pc;

  // inicializacia datoveho bloku
  memset(pWP, 0, sizeof(WORK_PARAMS));
  pWP->iSampleRateType = SAMPLE_RATE_MIDDLE;
  pWP->btAmplitude = ON_STATE;
  pWP->iBlockFrom = -1;
  pWP->iBlockTo = -1;
  pWP->iSilentGap = SILENT_GAP;

  // spracovanie a kontrola vstupnych parametrov
  do {
    for (i = 1; i < argc; i++) {
      if (strcmp(argv[i], "-h") == 0) {
        iErr = ERR_PRINT_USAGE;
        break;
      }
      else if (strcmp(argv[i], "-v") == 0) {
        iErr = ERR_PRINT_VERSION;
        break;
      }
      else if (strcmp(argv[i], "-vb") == 0) {
        pWP->bVB = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-mato") == 0) {
        pWP->bMato = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-dbl") == 0) {
        pWP->bDoubleSpeed = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-rph") == 0) {
        pWP->bRevPhase = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-list") == 0) {
        pWP->bList = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-blk") == 0) {
        if (i + 2 >= argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        else {
          i++;
          // pociatocny blok
          iV = parseNumber(argv[i]);
          if (iV < 1) { // test na maximalne cislo bloku je v `tape.c:loadTapeBlock()`
            iErr = printError(ERR_BAD_VALUE, argv[i], argv[i - 1]);
            break;
          }
          pWP->iBlockFrom = iV;

          i++;
          // posledny blok
          iV = parseNumber(argv[i]);
          if (iV < pWP->iBlockFrom) { // test na maximalne cislo bloku je v `tape.c:loadTapeBlock()`
            iErr = printError(ERR_BAD_VALUE, argv[i], argv[i - 1]);
            break;
          }
          pWP->iBlockTo = iV;
        }
      }
      else if (strcmp(argv[i], "-amp") == 0) {
        if (i + 1 >= argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        else {
          i++;
          // amplituda
          iV = parseNumber(argv[i]);
          if (iV < 1 || iV > 127) {
            iErr = printError(ERR_BAD_VALUE, argv[i], argv[i - 1]);
            break;
          }
          pWP->btAmplitude = (BYTE) (iV + 128);
        }
      }
      else if (strcmp(argv[i], "-gap") == 0) {
        if (i + 1 >= argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        else {
          i++;
          // medzera pred dalsim hlavickovym blokom
          iV = parseNumber(argv[i]);
          if (iV < 1) {
            iErr = printError(ERR_BAD_VALUE, argv[i], argv[i - 1]);
            break;
          }
          pWP->iSilentGap = iV;
        }
      }
      else if (strcmp(argv[i], "-sr") == 0) {
        if (i + 1 >= argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        else {
          i++;
          // sample rate
          if (stricmp(argv[i], "L") == 0)
            pWP->iSampleRateType = SAMPLE_RATE_LOW;
          else if (stricmp(argv[i], "M") == 0)
            pWP->iSampleRateType = SAMPLE_RATE_MIDDLE;
          else if (stricmp(argv[i], "H") == 0)
            pWP->iSampleRateType = SAMPLE_RATE_HIGH;
          else {
            iErr = printError(ERR_BAD_VALUE, argv[i], argv[i - 1]);
            break;
          }
        }
      }
      else if (strcmp(argv[i], "-o") == 0) {
        if (i + 1 == argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        if (pWP->pFnOut != NULL) {
          iErr = printError(ERR_OUTPUT_FILE_MORE, pWP->pFnIn, argv[i]);
          break;
        }
        i++;
        pWP->pFnOut = argv[i];
        continue;
      }
      else {
        if (*argv[i] == '-') {
          iErr = printError(ERR_UNKNOWN_OPTION, argv[i], NULL);
          break;
        }
        if (pWP->pFnIn != NULL) {
          iErr = printError(ERR_INPUT_FILE_MORE, pWP->pFnIn, argv[i]);
          break;
        }
        pWP->pFnIn = argv[i];
        continue;
      }
    }
    if (iErr != 0)
      break;

    if (pWP->pFnIn == NULL) {
      iErr = printError(ERR_NO_INPUT_FILE, NULL, NULL);
      break;
    }

    // priprava mena vystupneho suboru, ak nebol zadany
    if (pWP->pFnOut == NULL) {
      strcpy(pWP->szFn, pWP->pFnIn);
      pc = strrchr(pWP->szFn, '.');
      if (pc != NULL)
        *pc = 0;
      strcat(pWP->szFn, ".wav");
      pWP->pFnOut = pWP->szFn;
      if (stricmp(pWP->pFnOut, pWP->pFnIn) == 0)
        strcat(pWP->pFnOut, ".wav");
    }

  } while (FALSE);

  if (pWP->bDoubleSpeed && pWP->bMato)
    iErr = printError(ERR_BAD_OPTIONS, NULL, NULL);

  return iErr;
}
//---------------------------------------------------------------------------
long setFilePointer(FILE *f, long lOffset, int iWhence)
{
  long lPos;

  if (fseek(f, lOffset, iWhence) != 0)
    return ERR_ACCESS_SRC_FILE;

  lPos = ftell(f);
  if (lPos < 0)
    return ERR_ACCESS_SRC_FILE;

  return lPos;
}
//---------------------------------------------------------------------------
int readFile(char *pszFileName, BYTE *pbtBuff, long lOffset, DWORD dwLength)
{
  int iErr = ERR_OK;
  long lPos;
  FILE *f;

  do {
    f = fopen(pszFileName, "rb");
    if (f == NULL) {
      iErr = printError(ERR_OPEN_SRC_FILE, pszFileName, NULL);
      break;
    }
    if (fseek(f, lOffset, SEEK_SET) != 0) {
      iErr = printError(ERR_ACCESS_SRC_FILE, pszFileName, NULL);
      break;
    }
    lPos = ftell(f);
    if (lPos != lOffset) {
      iErr = printError(ERR_ACCESS_SRC_FILE, pszFileName, NULL);
      break;
    }
    if (fread(pbtBuff, 1, dwLength, f) != dwLength) {
      iErr = printError(ERR_READ_SRC_FILE, pszFileName, NULL);
      break;
    }
  } while (FALSE);

  if (f != NULL)
    fclose(f);

  return iErr;
}
//---------------------------------------------------------------------------
int allocBuffer(void **ppBuff, int iSize)
{
  if (*ppBuff)
    return ERR_OK;

  *ppBuff = (void *) malloc(iSize);
  if (!*ppBuff)
    return printError(ERR_OUT_OF_MEMORY, NULL, NULL);

  memset(*ppBuff, 0, iSize);
  return ERR_OK;
}
//---------------------------------------------------------------------------
int parseNumber(char *pszStr)
{
  char *p;
  int iVal = strtol(pszStr, &p, 0);
  if (iVal < 0 || (iVal == 0 && *p != 0))
    iVal = -1;
  return iVal;
}
//---------------------------------------------------------------------------
int printError(int iErr, void *pPar, void *pPar2)
{
  switch (iErr) {
    case ERR_OUT_OF_MEMORY :
      fprintf(stderr, "Out of memory!");
      break;

    case ERR_BAD_OPTIONS :
      fprintf(stderr, "Bad combination of options!");
      break;

    case ERR_UNKNOWN_OPTION :
      fprintf(stderr, "Unknown option '%s'!", (char *) pPar);
      break;

    case ERR_NO_OPTION_VALUE :
      fprintf(stderr, "Missing option value(s) '%s'!", (char *) pPar);
      break;

    case ERR_OPEN_SRC_FILE :
      fprintf(stderr, "Can't open input file '%s'!", (char *) pPar);
      break;

    case ERR_ACCESS_SRC_FILE :
      fprintf(stderr, "Can't access input file '%s'!", (char *) pPar);
      break;

    case ERR_LENGTH_SRC_FILE :
      fprintf(stderr, "Can't retrieve length of the input file '%s'!", (char *) pPar);
      break;

    case ERR_READ_SRC_FILE :
      fprintf(stderr, "Can't read from the input file '%s'!", (char *) pPar);
      break;

    case ERR_OPEN_DEST_FILE :
      fprintf(stderr, "Can't open output file '%s'!", (char *) pPar);
      break;

    case ERR_ACCESS_DEST_FILE :
      fprintf(stderr, "Can't access output file '%s'!", (char *) pPar);
      break;

    case ERR_WRITE_DEST_FILE :
      fprintf(stderr, "Can't write to the output file '%s'!", (char *) pPar);
      break;

    case ERR_NO_INPUT_FILE :
      fprintf(stderr, "Input file not specified!");
      break;

    case ERR_INPUT_FILE_MORE :
      fprintf(stderr, "Only one input file can be specified! '%s' or '%s' ?", (char *) pPar, (char *) pPar2);
      break;

    case ERR_OUTPUT_FILE_MORE :
      fprintf(stderr, "Only one output file can be specified! '%s' or '%s' ?", (char *) pPar, (char *) pPar2);
      break;

    case ERR_BAD_VALUE :
      fprintf(stderr, "Bad value '%s' in option `%s` !", (char *) pPar, (char *) pPar2);
      break;

    case ERR_TAPE_CORRUPTED :
      fprintf(stderr, "Tape '%s' corrupted at file offset %08X", (char *) pPar, *(long *) pPar2);
      break;

    case ERR_BAD_BLOCK_NUMBER :
      fprintf(stderr, "Bad block number: %d", *(int *) pPar);
      break;

    default :
      fprintf(stderr, "Unknown error code: %d", iErr);
      break;
  }

  fprintf(stderr, "\n\n");

  return iErr;
}
//---------------------------------------------------------------------------
void version(void)
{
  printf("%s v%d.%d - %s\n",
         PROG_NAME, PROG_VERSION / 16, PROG_VERSION % 16, PROG_AUTHOR);
}
//---------------------------------------------------------------------------
void pressEnter(void)
{
  printf("\nPress ENTER ...");
  while (getchar() != '\n')
    ;
}
//---------------------------------------------------------------------------
void usage(void)
{
  version();

  printf("Usage: %s [options] infile.ext\n", PROG_EXE_NAME);
  printf("\nCommon options:\n");
  printf("  -o outfile.ext  output file name\n");
  printf("  -list\t\t  print tape content\n");
  printf("  -blk FROM TO\t  range of blocks to export [all]\n");
  printf("  -mato\t\t  export using Mato protocol\n");
  printf("  -dbl\t\t  double speed (only PMD 85)\n");
  printf("  -rph\t\t  reversal phase\n");
  printf("  -sr SR\t  sample rate [M]; SR=L/M/H (Low/Middle/High)\n");
  printf("  -amp NUM\t  amplitude <1, 127> [90]\n");
  printf("  -gap NUM\t  gap before next header block in seconds [4]\n");
  printf("  -vb\t\t  verbose\n");
  printf("  -h\t\t  usage information\n");
  printf("  -v\t\t  version info\n");

  printf("\nAll numeric parameters must be nonnegative integers and can be in three forms:\n");
  printf("  decimal (e.g. 12345)\n");
  printf("  octal (e.g. 01234)\n");
  printf("  hexadecimal (e.g. 0x1234)\n\n");
}
//---------------------------------------------------------------------------

